/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceAccessorBase_h__
#define __surface_SurfaceAccessorBase_h__

namespace fe
{

/**************************************************************************//**
    @brief Common Functionality for Accessor Surface

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessorBase:
	virtual public SurfaceAccessorI
{
	public:
						SurfaceAccessorBase(void)
						{	setName("SurfaceAccessorBase"); }
virtual					~SurfaceAccessorBase(void)							{}

						//* as SurfaceAccessorI
virtual	String			type(void) const				{ return m_attrType; }
virtual	U32				count(void) const;
virtual	U32				subCount(U32 a_index) const;

virtual	void			set(U32 a_index,U32 a_subIndex,String a_string);
virtual	void			set(U32 a_index,String a_string)
						{	set(a_index,U32(0),a_string); }
virtual	String			string(U32 a_index,U32 a_subIndex=0);

virtual	void			set(U32 a_index,U32 a_subIndex,I32 a_integer);
virtual	void			set(U32 a_index,I32 a_integer)
						{	set(a_index,U32(0),a_integer); }
virtual	I32				integer(U32 a_index,U32 a_subIndex=0);

virtual	I32				duplicate(U32 a_index,U32 a_subIndex=0)	{ return -1; }

virtual	I32				append(void);
virtual	void			append(U32 a_index,I32 a_integer);
virtual	I32				append(I32 a_integer);
virtual	void			append(Array< Array<I32> >& a_rPrimVerts);

virtual	void			remove(U32 a_index,I32 a_integer=0)					{}

virtual	void			set(U32 a_index,U32 a_subIndex,Real a_real);
virtual	void			set(U32 a_index,Real a_real)
						{	set(a_index,U32(0),a_real); }
virtual	Real			real(U32 a_index,U32 a_subIndex=0);

virtual	void			set(U32 a_index,U32 a_subIndex,
								const SpatialVector& a_vector);
virtual	void			set(U32 a_index,const SpatialVector& a_vector)
						{	set(a_index,U32(0),a_vector); }
virtual	SpatialVector	spatialVector(U32 a_index,U32 a_subIndex=0);

virtual	void			fragmentWith(SurfaceAccessibleI::Element a_element,
								String a_attributeName,String a_groupName);
virtual	U32				fragmentCount(void) const
						{	return m_fragmentMap.size(); }
virtual	String			fragment(U32 a_index) const
						{
							AutoHashMap<String,Fragment>::const_iterator it=
									m_fragmentMap.begin();
							for(U32 m=0;m<a_index;m++)
							{
								if(it==m_fragmentMap.end())
								{
									return String();
								}
								it++;
							}
							if(it==m_fragmentMap.end())
							{
								return String();
							}
							return it->first;
						}
virtual	BWORD			filterWith(String a_filterString,
							sp<SurfaceAccessibleI::FilterI>& a_rspFilter) const
						{
							AutoHashMap<String,Fragment>::const_iterator it=
									m_fragmentMap.find(a_filterString);
							if(it==m_fragmentMap.end())
							{
								a_rspFilter=NULL;
								return FALSE;
							}
							sp<FragmentFilter> spFragmentFilter=a_rspFilter;
							if(spFragmentFilter.isNull())
							{
								spFragmentFilter=sp<FragmentFilter>(
										new FragmentFilter());
								FEASSERT(spFragmentFilter.isValid());
							}
							spFragmentFilter->setFragment(&(it->second));
							a_rspFilter=spFragmentFilter;
							return TRUE;
						}

		void			setSurfaceAccessible(
								sp<SurfaceAccessibleI> a_spSurfaceAccessibleI)
						{	m_spSurfaceAccessibleI=a_spSurfaceAccessibleI; }

virtual	SurfaceAccessibleI::Element		element(void)	{ return m_element; }
virtual	SurfaceAccessibleI::Attribute	attribute(void)	{ return m_attribute; }

virtual	String			attributeName(void)		{ return m_attrName; }

virtual	void			appendGroupSpans(
								SpannedRange::MultiSpan& a_rMultiSpan);

		BWORD			bind(SurfaceAccessibleI::Element a_element,
								SurfaceAccessibleI::Attribute a_attribute)
						{
							m_attribute=a_attribute;

							String name;
							switch(a_attribute)
							{
								case SurfaceAccessibleI::e_generic:
								case SurfaceAccessibleI::e_position:
									name="spc:at";
									break;
								case SurfaceAccessibleI::e_normal:
									name="spc:up";
									break;
								case SurfaceAccessibleI::e_uv:
									name="surf:uvw";
									break;
								case SurfaceAccessibleI::e_vertices:
									name="surf:verts";
									break;
//									m_attrName="vertices";
//									FEASSERT(a_element==
//											SurfaceAccessibleI::e_primitive);
//									m_element=a_element;
//									return TRUE;
								case SurfaceAccessibleI::e_properties:
									name="properties";
									break;
//									m_attrName="properties";
//									FEASSERT(a_element==
//											SurfaceAccessibleI::e_primitive);
//									m_element=a_element;
//									return TRUE;
							}
							return bindInternal(a_element,name);
						}
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							String name=a_name;
							m_attribute=SurfaceAccessibleI::e_generic;

							if(name=="P")
							{
								name="spc:at";
								m_attribute=SurfaceAccessibleI::e_position;
							}
							else if(name=="N")
							{
								name="spc:up";
								m_attribute=SurfaceAccessibleI::e_normal;
							}
							else if(name=="uv")
							{
								name="surf:uvw";
								m_attribute=SurfaceAccessibleI::e_uv;
							}

							return bindInternal(a_element,name);
						}

		void			bind(WeakRecord& a_rRecord)
						{	m_weakRecord=a_rRecord; }

	protected:

		SurfaceAccessibleI::Element		m_element;
		SurfaceAccessibleI::Attribute	m_attribute;
		String							m_attrName;
		String							m_attrType;
		sp<SurfaceAccessibleI>			m_spSurfaceAccessibleI;

	private:

virtual	BWORD			bindInternal(SurfaceAccessibleI::Element a_element,
								const String& a_name);

		sp<RecordArray>	recordArray(void) const;
		void			ensureType(String a_attrType);

	class Fragment
	{
		public:
			Array<U32>	m_indexArray;
	};
	class FragmentFilter: public SurfaceAccessibleI::FilterI
	{
		public:
					FragmentFilter(void)									{}
	virtual			~FragmentFilter(void)									{}

	virtual	I32		filterCount(void) const
					{
						return m_pFragment?
								m_pFragment->m_indexArray.size(): -1;
					}
	virtual	I32		filter(I32 a_filterIndex) const
					{
						if(!m_pFragment)
						{
							return -1;
						}
						const Array<U32>& rIndexArray=
								m_pFragment->m_indexArray;
						if(a_filterIndex>=I32(rIndexArray.size()))
						{
							return -1;
						}
						return rIndexArray[a_filterIndex];
					}

					/// @internal
			void	setFragment(const Fragment* a_pFragment)
					{	m_pFragment=a_pFragment; }

		private:
			const Fragment*			m_pFragment;
	};

		WeakRecord						m_weakRecord;
		sp<RecordGroup>					m_spRG;
		sp<RecordArray>					m_spRecordArray;
		sp<Layout>						m_spLayout;
		SurfacePoint					m_surfaceElementRV;
		SurfacePoint					m_surfacePointRV;	//* for e_vertices

		Accessor<String>				m_accessorString;
		Accessor<I32>					m_accessorInteger;
		Accessor<Real>					m_accessorReal;
		Accessor<SpatialVector>			m_accessorSpatialVector;

		AutoHashMap<String,Fragment>	m_fragmentMap;
		Accessor< Array<I32> >			m_vertexIndexAccessor;
};

} // namespace

#endif /* __surface_SurfaceAccessorBase_h__ */
